=pod

---+ package XmlRpcPlugin

Make XML RPC calls to remote web services.

=cut

use strict;
package TWiki::Plugins::XmlRpcPlugin;
our $pluginName = 'XmlRpcPlugin';
require TWiki::Func;
require TWiki::Plugins;
our $VERSION = '$Rev: 21319 (2012-05-25) $';
our $RELEASE = '2012-05-22';
our $SHORTDESCRIPTION = 'This plugin implements XML RPC calls through the %XMLRPC{}% variable';
our $NO_PREFS_IN_TOPIC = 1;

# Define other global package variables
our $debug;
our $cache;
our $refresh;
our $default_url;
our $domains_pattern;
our $json;
our $default_timeout;
our $default_duration;

=pod

---++ initPlugin($topic, $web, $user, $installWeb) -> $boolean
   * =$topic= - the name of the topic in the current CGI query
   * =$web= - the name of the web in the current CGI query
   * =$user= - the login name of the user
   * =$installWeb= - the name of the web the plugin is installed in

=cut

sub initPlugin {
    my( $topic, $web, $user, $installWeb ) = @_;
    if( $TWiki::Plugins::VERSION < 1.1 ) {
        TWiki::Func::writeWarning( "Version mismatch between $pluginName and Plugins.pm" );
        return 0;
    }
    TWiki::Func::registerTagHandler( 'XMLRPC', \&_XMLRPC );
    return 1;
}
sub TO_JSON { return { "hi" }; }
# The function used to handle the %XMLRPC{...}% variable
sub _XMLRPC {
    my($session, $params, $theTopic, $theWeb, $meta, $textRef) = @_;
    # $session  - a reference to the TWiki session object (if you don't know
    #             what this is, just ignore it)
    # $params=  - a reference to a TWiki::Attrs object containing parameters.
    #             This can be used as a simple hash that maps parameter names
    #             to values, with _DEFAULT being the name for the default
    #             parameter.
    # $theTopic - name of the topic in the query
    # $theWeb   - name of the web in the query
    # $meta     - topic meta-data to use while expanding, can be undef (Since TWiki::Plugins::VERSION 1.4)
    # $textRef  - reference to unexpanded topic text, can be undef (Since TWiki::Plugins::VERSION 1.4)

    # if debug set in the config, write to debug.txt; if debug is set in preferences, write to the topic
    $debug = TWiki::Func::getPreferencesFlag( "\U$pluginName\E_DEBUG" ) || $TWiki::cfg{Plugins}{XmlRpcPlugin}{Debug} if !defined $debug;
    require Data::Dumper;
    my $debugMessage;
    my $result;

    eval {
      $debugMessage .= "XMLRPC: Received from topic:\n\n" . Data::Dumper->Dumper($params->{_RAW}) . "\n\n" if $debug;

      # check if the required condition is satisfied otherwise return
      !defined $params->{requires} || $params->{requires} || die "required condition not satistfied\n";

      require URI;
      my $url;
      $domains_pattern = "(" . join("\$|", split(/ +/, TWiki::Func::getPreferencesValue( "\U$pluginName\E_DOMAINS" ))) . "\$)" if !defined $domains_pattern;
      if(defined $params->{url}){
        $url = $params->{url};
        if(URI->new($url, "http")->host !~ m/$domains_pattern/i){ die "url=\"" . $url . "\" does not match allowed domains specified by \U$pluginName\E_DOMAINS, pattern=" . $domains_pattern . "\n"; }
      }else{
        if(!defined $default_url){
          $default_url = TWiki::Func::getPreferencesValue( "\U$pluginName\E_URL" ) || "";
          if(URI->new($default_url, "http")->host !~ m/$domains_pattern/i){ 
            $default_url = ""; #invalidate default url
            die "\U$pluginName\E_URL=\"" . TWiki::Func::getPreferencesValue( "\U$pluginName\E_URL" ) . "\" does not match allowed domains specified by \U$pluginName\E_DOMAINS, pattern=" . $domains_pattern . "\n";
          }
        }
        $url = $default_url;
      }
      $url eq "" && die "url is empty or not allowed\n";

      $params->{_DEFAULT} =~ /^([\w.]+)\((.*)\)$/ || die "Function call must be in the format: \"function(JSON_par1, JSON_par2, ...)\"\n";
      my $function = $1;
      my $json_params = "[" . $2 . "]";

      my $key = $params->{_DEFAULT} . $url;
      require Time::Duration::Parse;
      $default_duration = Time::Duration::Parse::parse_duration(TWiki::Func::getPreferencesValue( "\U$pluginName\E_EXPIRES_IN" ) || "1y") if !defined $default_duration;
      my $duration = defined $params->{expires_in} ? Time::Duration::Parse::parse_duration($params->{expires_in}) : $default_duration;
      $debugMessage .= "XMLRPC: Calculated duration in seconds: " . $duration . "\n\n" if $debug;
      $refresh = TWiki::Func::getCgiQuery()->param('xmlrpc') || 0 if !defined $refresh;
      require CHI;
      $cache = CHI->new(
        driver     => 'File',
        root_dir   => $TWiki::cfg{Plugins}{XmlRpcPlugin}{CachePath} || TWiki::Func::getWorkArea($pluginName),
        namespace => $theWeb
      ) if !defined $cache;
      require JSON::PP;
      $json = JSON::PP->new->allow_singlequote->allow_barekey->allow_blessed->convert_blessed if !defined $json;
      my $rawresult = $refresh ? undef : $cache->get($key, expire_if => sub { time() - $_[0]->created_at > $duration });
      if(defined $rawresult){
        $debugMessage .= "XMLRPC: Got result from cache:\n\n" . Data::Dumper->Dumper($rawresult) . "\n\n" if $debug;
      } else {
        my $parameters = $json->decode($json_params);

        $debugMessage .= "XMLRPC: Making call to server:\n\n" . Data::Dumper->Dumper($url) . "\n\nXMLRPC: Making call with function:\n\n" . Data::Dumper->Dumper($function) . "\n\nXMLRPC: Making call with parameters:\n\n" . Data::Dumper->Dumper($parameters) . "\n\n" if $debug;
        eval {
          require Sys::SigAction;
          require XML::RPC;
          $default_timeout = TWiki::Func::getPreferencesValue( "\U$pluginName\E_TIMEOUT" ) || 10 if !defined $default_timeout;
          my $timeout = defined $params->{timeout} ? $params->{timeout} : $default_timeout;
          Sys::SigAction::timeout_call($timeout, sub{ $rawresult = XML::RPC->new($url)->call($function, @$parameters); });
          1;
        } or do {
          my $exception = $@;
          $debugMessage .= "XMLRPC: Exception in connecting to server:\n\n" . Data::Dumper->Dumper($exception) . "\n\n" if $debug;
          die "Connection timed out\n"; 
        };
        $debugMessage .= "XMLRPC: Received from server:\n\n" . Data::Dumper->Dumper($rawresult) . "\n\n" if $debug;

        $cache->set($key, $rawresult, defined $params->{expires_in} ? $params->{expires_in} : "never");
      }

      # select the desired part of the object
      my @fields = split /\./, $params->{select};
      my $selectedresult = $rawresult;
      foreach my $word (@fields) {
        next if $word eq "";
        if (ref $selectedresult eq "ARRAY" && $word =~ /\[(\d+)\]/) {
          $selectedresult = $selectedresult->[$1];
        } elsif (ref $selectedresult eq "HASH") {
          $selectedresult = $selectedresult->{$word};
        }
      }
      $debugMessage .= "XMLRPC: Selected result:\n\n" . Data::Dumper->Dumper($selectedresult) . "\n\n" if $debug;

      my $class = ref $selectedresult;
      if($class eq "") {
        $result = $selectedresult;
      } else {
        # return objects in JSON format
        my $json_answer = $json->encode($selectedresult);
        # fixme: doesn't work when there are quotes in the values
        $json_answer =~ s/"/'/g;
        $result = $json_answer;
      }
      1;
    } or do {
      if($debug){
        $result = undef;
        my $exception = $@;
        $debugMessage .= "XMLRPC Exception:\n\n" . Data::Dumper->Dumper( $exception ) . "\n\n";
      }
    };

    # return
    if(!defined $result){ $result = defined $params->{"if_fails"} ? $params->{"if_fails"} : ""; }
    $debugMessage .= "XMLRPC: Returning to topic:\n\n" . $result . "\n\n" if $debug;
    TWiki::Func::writeDebug( $debugMessage ) if $debug && $TWiki::cfg{Plugins}{XmlRpcPlugin}{Debug};
    return ($debug && TWiki::Func::getPreferencesFlag( "\U$pluginName\E_DEBUG" )) ? $debugMessage : $result;
}

1;
